{
	Code Linguine Source Code Parsing/Query Library
	Copyright (C) 2000-2005 Pavel Simakov
	http://www.softwaresecretweapons.com

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
}

{$I oycl.inc}  

unit OYDelphi5Preproc;

interface

uses { vcl } SysUtils, Classes,
     { ly  } LexLib, OYDelphi5Lexer, OYDelphi5Parser, OYParserToken;

type
  { list of supported compiler directives }
  TOYPascalCompilerDirectiveKind = (
    dkDefine, dkElse, dkEndif, dkIfdef, dkIfndef,
    dkIfopt, dkInclude, dkUndef, dkNull
  );

  { generic preprocessor }
  TOYDelphi5Preproc = class(TObject)
  private
    fUrl : String;
    fStream : TStream;
    fTokens : TList;
    fReady : Boolean;
    fCurrent : Integer;
    fDefines : TStrings;
    fBuiltIn : TStrings;
    fLexers : TList;
    fOnWarning : TOYParseEvent;
    fOnHint : TOYParseEvent;
    fOnCanContinue : TOYParserCanContinue;

  private
    procedure AppendTokenListFromLexer(Lexer : TLexer; Index : Integer);
    procedure ScanForDirective(Parent : TOYPascalCompilerDirectiveKind;  ParentIgnore : Boolean);
    function GetDirectiveKind(Directive : String) : TOYPascalCompilerDirectiveKind;
    function GetDirectiveParameter(Directive : String) : String;
    procedure ProcessInclude(const Url : String);

    procedure Warning(URL, reason, line, token : WideString; lineno, colno : Integer);
    procedure Hint(URL, reason, line, token : WideString; lineno, colno : Integer);
    function CanContinue : Boolean;


  public
    constructor Create(const Url : String);
    destructor Destroy; override;

  public
    function Lex : TOYParserToken;
    function CurrentToken : POYParserToken;
    procedure FirstToken;
    procedure NextToken;
    procedure PrevToken;
    procedure LastToken;
    function lineByToken(Token : TOYParserToken) : String;
    function textByToken(Token : TOYParserToken) : String;
    function urlByToken(Token : TOYParserToken) : String;

  public
    property OnWarning : TOYParseEvent read fOnWarning write fOnWarning;
    property OnHint : TOYParseEvent read fOnHint write fOnHint;
    property OnCanContinue : TOYParserCanContinue read fOnCanContinue write fOnCanContinue;

  end;

implementation

{ TOYDelphi5Preproc }
constructor TOYDelphi5Preproc.Create(const Url : String);
begin
  inherited Create;

  fUrl := Url;
  fReady := False;                                                      
  fCurrent := 0;
  
  fStream := TFileStream.Create(Url, fmOpenRead + fmShareDenyWrite);
  fTokens := TList.Create;

  { define buil ins }
  fDefines := TStringList.Create;

  fBuiltIn := TStringList.Create;
  fBuiltIn.Add('VER130');
  fBuiltIn.Add('WIN32');
  fBuiltIn.Add('CPU386');

  fLexers := TList.Create;
end;

destructor TOYDelphi5Preproc.Destroy;
var
  i : Integer;
begin
  for i :=0 to fTokens.Count - 1 do begin
    Dispose(POYParserToken(fTokens[i]));
  end;

  for i := 0 to fLexers.Count - 1 do begin
    TObject(fLexers[i]).Free;
  end;
  
  fLexers.Free;
  fBuiltIn.Free;
  fDefines.Free;
  fTokens.Free;
  fStream.Free;
  inherited Destroy;
end;

procedure TOYDelphi5Preproc.Warning(URL, reason, line, token : WideString; lineno, colno : Integer);
begin
  if Assigned(fOnWarning) then begin
    fOnWarning(Url, Reason, Line, Token, Lineno, Colno);
  end;
end;

procedure TOYDelphi5Preproc.Hint(URL, reason, line, token : WideString; lineno, colno : Integer);
begin
  if Assigned(fOnHint) then begin
    fOnHint(Url, Reason, Line, Token, Lineno, Colno);
  end;
end;

function TOYDelphi5Preproc.CanContinue : Boolean;
begin
  Result := True;
  if Assigned(fOnCanContinue) then begin
    Result := fOnCanContinue;
  end;
end;

procedure TOYDelphi5Preproc.AppendTokenListFromLexer(Lexer : TLexer; Index : Integer);
var
  aRes : Integer;
  pToken : POYParserToken;
begin
  repeat
    aRes := Lexer.lex;
    if aRes <> 0 then begin
      New(pToken);
      pToken^.Lexer := Lexer;
      pToken^.Frame := Lexer.frame;
      fTokens.Insert(Index, pToken);
      Inc(Index);
    end else begin
      break;
    end;

    if ((Index mod 100) = 0) and (not CanContinue) then begin
      raise Exception.Create('Processing aborted.');
    end;

  until False;

  if fTokens.Count = 0 then begin
    raise Exception.Create('No tokens found.');
  end;
end;

function TOYDelphi5Preproc.CurrentToken : POYParserToken;
begin
  Result := nil;
  if (fTokens.Count > 0) and (fCurrent < fTokens.Count) then begin
    Result := fTokens[fCurrent];
  end;
end;

procedure TOYDelphi5Preproc.FirstToken;
begin
  fCurrent := 0;
end;

procedure TOYDelphi5Preproc.LastToken;
begin
  fCurrent := fTokens.Count;
end;

procedure TOYDelphi5Preproc.NextToken;
begin
  Inc(fCurrent);
  if fCurrent > fTokens.Count then begin
    fCurrent := fTokens.Count;
  end;
end;

procedure TOYDelphi5Preproc.PrevToken;
begin
  Dec(fCurrent);
  if fCurrent < 0 then begin
    fCurrent := 0;
  end;
end;

function TOYDelphi5Preproc.Lex : TOYParserToken;
var
  aLexer : TLexer;
begin
  if not fReady then begin

    aLexer := TTemplatedLexer.Create(fUrl);
    aLexer.OnWarning := Warning;
    aLexer.OnHint := Hint;
    aLexer.OnCanContinue := CanContinue;
    fLexers.Add(aLexer);

    AppendTokenListFromLexer(aLexer, 0);
    ScanForDirective(dkNull, False);
    fCurrent := 0;
    fReady := True;
  end;

  FillChar(Result, SizeOf(Result), 0);
  if (fTokens.Count > 0) and (fCurrent < fTokens.Count) then begin
    Result := POYParserToken(fTokens[fCurrent])^;
    NextToken;
  end;
end;                               

function TOYDelphi5Preproc.GetDirectiveParameter(Directive : String) : String;
var
  aLeft, aRight : Integer;
begin
  aLeft := Pos(' ', Directive);
  if aLeft = 0 then begin
    Result := Directive;
  end else begin
    aRight := Pos('}', Directive);
    if aRight = 0 then begin
      aRight := Length(Directive);  
    end;
    Result := UpperCase(Copy(Directive, aLeft + 1, aRight - aLeft - 1));
  end;          
end;

function TOYDelphi5Preproc.GetDirectiveKind(Directive : String) : TOYPascalCompilerDirectiveKind;
var
  aLeft, aRight, aSpace, aBracket : Integer;
  aDirective : String;
begin
  aLeft := Pos('{', Directive);
  aSpace := Pos(' ', Directive);
  aBracket := Pos('}', Directive);

  if aBracket = 0 then begin        
    aBracket := Length(Directive);
  end;

  if (aSpace > 0) and (aSpace < aBracket) then begin
    aRight := aSpace;
  end else begin
    aRight := aBracket;
  end;

  if (aLeft <> 0) and (aRight <> 0) then begin
    aDirective := UpperCase(Copy(Directive, aLeft + 1, aRight - aLeft - 1));
  end else begin
    aDirective := '';
  end;

  if aDirective = '$IFDEF' then begin Result := dkIfdef; end else begin
  if aDirective = '$IFNDEF' then begin Result := dkIfndef; end else begin
  if aDirective = '$IFOPT' then begin Result := dkIfopt; end else begin
  if aDirective = '$I' then begin Result := dkInclude; end else begin
  if aDirective = '$ELSE' then begin Result := dkElse; end else begin
  if aDirective = '$ENDIF' then begin Result := dkEndif; end else begin
  if aDirective = '$DEFINE' then begin Result := dkDefine; end else begin
  if aDirective = '$UNDEF' then begin Result := dkUndef; end else begin
    Result := dkNull;
  end end end end end end end end;
end;

procedure TOYDelphi5Preproc.ProcessInclude(const Url : String);
var
  aLexer : TTemplatedLexer;
begin
  aLexer := TTemplatedLexer.Create(ExtractFilePath(fUrl) + Url);
  aLexer.OnWarning := Warning;
  aLexer.OnHint := Hint;
  aLexer.OnCanContinue := CanContinue;
  fLexers.Add(aLexer);
  
  AppendTokenListFromLexer(aLexer, fCurrent);
end;

procedure TOYDelphi5Preproc.ScanForDirective(Parent :TOYPascalCompilerDirectiveKind;  ParentIgnore : Boolean);

  function DirectiveKindToString(Directive : TOYPascalCompilerDirectiveKind) : String;
  begin
    case Directive of
      dkDefine : Result := '$DEFINE';
      dkElse : Result := '$ELSE';
      dkEndif : Result := '$ENDIF';
      dkIfdef : Result := '$IFDEF';
      dkIfndef : Result := '$IFNDEF';
      dkIfopt : Result := '$IFOPT';
      dkInclude : Result := '$INCLUDE';
      dkUndef : Result := '$UNDEF';
      else raise Exception.Create('Unknown compiler directive kind(' + IntToStr(Integer(Directive)) + ').');
    end;

  end;


  procedure _def(Name : String);
  begin
    if fDefines.IndexOf(Name) = -1 then begin
      fDefines.Add(Name);
    end;
  end;

  procedure _undef(Name : String);
  begin
    if fDefines.IndexOf(Name) <> -1 then begin
      fDefines.Delete(fDefines.IndexOf(Name));
    end;
  end;

  function _ifdef(Name : String) : Boolean;
  begin
    Result := (fDefines.IndexOf(Name) <> -1) or (fBuiltIn.IndexOf(Name) <> -1);
  end;

var
  aKind, aLastKind : TOYPascalCompilerDirectiveKind;
  aParameter : String;
  aIgnore : Boolean;
begin
  aIgnore := ParentIgnore;  
  aLastKind := dkNull;
  while CurrentToken <> nil do begin

    case CurrentToken.Frame.id of
      Integer(tkCompiler) : begin

        aKind := GetDirectiveKind(textByToken(CurrentToken^));
        aParameter := GetDirectiveParameter(textByToken(CurrentToken^));
                                           
        case aKind of
          dkInclude : begin
            if not aIgnore then begin
              NextToken;
              ProcessInclude(aParameter);
              continue;
            end;
          end;

          dkDefine : begin
            if not aIgnore then begin
              _def(aParameter);
            end;
          end;

          dkUndef : begin
            if not aIgnore then begin
              _undef(aParameter);
            end;
          end;

          dkIfdef, dkIfndef, dkIfopt : begin
            if aKind = dkIfopt then begin
              Warning(urlByToken(CurrentToken^), 'IFOPT is not fully supported (always evaluates to TRUE).', textByToken(CurrentToken^), lineByToken(CurrentToken^), CurrentToken^.Frame.LeftLineIndex, CurrentToken^.Frame.LeftColIndex);
            end;
            if (aLastKind in [dkIfdef, dkIfndef, dkIfopt, dkElse]) then begin
              ScanForDirective(aKind, aIgnore);
            end else begin
              aLastKind := aKind;
              if not aIgnore then begin  
                case aKind of
                  dkIfdef : aIgnore := not _ifdef(aParameter);
                  dkIfndef : aIgnore := _ifdef(aParameter);
                  dkIfopt : aIgnore := False;
                end;
              end;
            end;
          end;

          dkElse : begin
            if not (aLastKind in [dkIfdef, dkIfndef, dkIfopt] ) then begin
              raise Exception.Create('Directive $ELSE before $IFDEF.');
            end;
            aLastKind := aKind;
            if not ParentIgnore then begin
              aIgnore := not aIgnore;
            end;
          end;

          dkEndif : begin
            if not (aLastKind in [dkIfdef, dkIfndef, dkIfopt, dkElse] ) then begin
              raise Exception.Create('Directive $ENDIF before $IFDEF.');
            end;
            aIgnore := False;    
            aLastKind := aKind;
            if Parent <> dkNull then begin
              break;
            end;
          end;

        end;
      end;

      else begin
        if aIgnore then begin
          CurrentToken.Frame.id := tkIgnore;
        end;
      end;

    end;                 
    NextToken;
  end;

  if not((aLastKind = dkEndif) or (aLastKind = dkNull)) then begin
    raise Exception.Create('Unterminated compiler directive (' + DirectiveKindToString(Parent) + ').');
  end;
end;

function TOYDelphi5Preproc.lineByToken(Token : TOYParserToken) : String;
begin
  Result := TLexer(Token.Lexer).lineByFrame(Token.Frame);
end;

function TOYDelphi5Preproc.textByToken(Token : TOYParserToken) : String;
begin
  Result := TLexer(Token.Lexer).textByFrame(Token.Frame);
end;

function TOYDelphi5Preproc.urlByToken(Token : TOYParserToken) : String;
begin
  Result := TLexer(Token.Lexer).url;
end;

end.
