{
	Code Linguine Source Code Parsing/Query Library
	Copyright (C) 2000-2005 Pavel Simakov
	http://www.softwaresecretweapons.com

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
}

{$I oycl.inc}  

unit OYDelphi5Producer;

interface

uses { vcl } SysUtils, Classes,
     { oy  } OYDelphi5Parser, OYDelphi5Preproc, OYParserToken, OYYaccLib,
             OYApplication, OYPackage, OYProcessingInfoF, LexLib, OYDelphi5Lexer;

type
  { TOYDelphi5Producer }
  TOYDelphi5Producer = class(TObject)
  private
    fIContext : IOYLinguineExecutionContext;
    fIConsumer : IOYLinguineConsumer;
    fIInfo : IOYLinguineProcessingInfo;
    procedure Warning(URL, reason, line, token : WideString; lineno, colno : Integer);
    procedure Hint(URL, reason, line, token : WideString; lineno, colno : Integer);
    function CanContinue : Boolean;
  public
    constructor Create(const IContext : IOYLinguineExecutionContext; const IConsumer : IOYLinguineConsumer);
    destructor Destroy; override;

  public
    function Execute(const Url : String) : Boolean;
    function GetProcessingInfo : IOYLinguineProcessingInfo;
  end;

  { TOYDelphi5Highlighter }
  TOYDelphi5Highlighter = class(TOYDelphi5SyntaxHighlighter)
  private
    fISchema : IOYLinguineSchema;
  protected
    function IHighlightSchema : IOYLinguineSchema; override;
    function GetTokenIndexInSchemaForFrame(Frame : TOYLexerFrame) : Integer; override;
  end;

  { TOYDelphi5TemplatedParser }
  TOYDelphi5TemplatedParser = class(TTemplatedParser)
  private
    fNeedOutput : Boolean;
    fParsed : Boolean;
    fPreproc : TOYDelphi5Preproc;
    fIConsumer : IOYLinguineConsumer;
    fTokenIndex : Integer;

  protected
    function yylex : TOYParserToken; override;

  public
    constructor Create(const Url : String; const IConsumer : IOYLinguineConsumer);
    destructor Destroy; override;

  public
    function yyparse : Boolean; override;
    procedure yyproduce;
    function currentToken : POYParserToken;
    function lineByToken(Token : TOYParserToken) : String; override;
    function textByToken(Token : TOYParserToken) : String; override;
    function urlByToken(Token : TOYParserToken) : String; override;
  end;

  TOYSchemaData = record
    Name : String;                             
    Desc : String;
  end;

  { TOYDelphi5Schema }
  TOYDelphi5Schema = class(TOYHighlightLinguineSchema)
  protected
    procedure ConfigureGroups; override;
  end;

implementation

uses { oy } OYUtility;

const
  { XML element names }
  cCRLFElementName = 'l';

{ TOYDelphi5Producer }
constructor TOYDelphi5Producer.Create(const IContext : IOYLinguineExecutionContext; const IConsumer : IOYLinguineConsumer);
begin
  inherited Create;
  fIContext := IContext;
  fIConsumer := IConsumer;
end;

destructor TOYDelphi5Producer.Destroy;
begin
  fIContext := nil;
  fIConsumer := nil;
  inherited Destroy;
end;

procedure TOYDelphi5Producer.Warning(URL, reason, line, token : WideString; lineno, colno : Integer);
begin
  fIInfo := TOYProcessingInfo.Create(icWarning, 1, Url, Reason, line, token, -1, Lineno, Colno, fIInfo);
end;

procedure TOYDelphi5Producer.Hint(URL, reason, line, token : WideString; lineno, colno : Integer);
begin
  fIInfo := TOYProcessingInfo.Create(icHint, 1, Url, Reason, line, token, -1, Lineno, Colno, fIInfo);
end;

function TOYDelphi5Producer.GetProcessingInfo : IOYLinguineProcessingInfo;
begin
  Result := fIInfo;
end;

function TOYDelphi5Producer.CanContinue : Boolean;
begin
  Result := fIContext.IProgress.CanContinueBusy;
end;

function TOYDelphi5Producer.Execute(const Url : String) : Boolean;
var
  aParser : TOYDelphi5TemplatedParser;
begin
  try
    aParser := TOYDelphi5TemplatedParser.Create(Url, fIConsumer);
    try
      aParser.OnWarning := Warning;
      aParser.OnHint := Hint;
      aParser.OnCanContinue := CanContinue;
      try
        Result := aParser.yyparse;
        if not Result then begin
          fIInfo := TOYProcessingInfo.Create(icError, 1, aParser.urlByToken(aParser.currentToken^), 'Syntax error.', aParser.lineByToken(aParser.currentToken^), '', -1, aParser.currentToken.Frame.leftLineIndex, aParser.currentToken.Frame.leftColIndex, fIInfo);
        end;
      except
        on E : Exception do begin
          Result := False;
          if (aParser.currentToken = nil) then begin
            raise;
          end;
          fIInfo := TOYProcessingInfo.Create(icError, 1, aParser.urlByToken(aParser.currentToken^), E.Message, aParser.lineByToken(aParser.currentToken^), '', -1, aParser.currentToken.Frame.leftLineIndex, aParser.currentToken.Frame.leftColIndex, fIInfo);
          Result := False;
        end;
      end;
      if Result then begin
        aParser.yyproduce;
      end;
    finally
      aParser.Free;
    end;
  except
    on E : Exception do begin
      fIInfo := TOYProcessingInfo.Create(icError, 1, Url, E.Message, '', '', -1, -1, -1, fIInfo);
      Result := False;
    end;
  end;
end;


{ TOYDelphi5Highlighter }
function TOYDelphi5Highlighter.IHighlightSchema : IOYLinguineSchema;
begin
  if not Assigned(fISchema) then begin
    fISchema := TOYDelphi5Schema.Create;
  end;
  Result := fISchema;
end;

function TOYDelphi5Highlighter.GetTokenIndexInSchemaForFrame(Frame : TOYLexerFrame) : Integer;
begin
  Result := 0;
  case Frame.Id of
    tkArray..tkWith : Result := 1;
    tkAsciiChar, tkStringLiteral : Result := 2;
    tkInteger, tkHex, tkFloat : Result := 3;
    tkAnsiComment..tkSlashesComment : Result := 4;
    tkAddressOf..tkSymbol, tkEqual..tkNotEqual, tkMinus..tkPlus : Result := 5;
    tkAsmBody : Result := 6;
    tkCompiler, tkRegister, tkPascal, tkCdecl, tkSafecall,
    tkNear, tkFar, tkStdcall, tkResident, tkExport : Result := 7;
  end;
end;

{ TOYDelphi5Schema }
procedure TOYDelphi5Schema.ConfigureGroups;
begin
  fGroups := 8;
  SetLength(fGroupNames, fGroups);
  fGroupNames[0].Name := 'Other';
  fGroupNames[0].Desc := 'All other language elements.';
  fGroupNames[1].Name := 'Reserved Words';
  fGroupNames[1].Desc := 'All reserved words.';
  fGroupNames[2].Name := 'String Literals';
  fGroupNames[2].Desc := 'All string literals.';
  fGroupNames[3].Name := 'Numbers';
  fGroupNames[3].Desc := 'All integer, float, and hexadecimal numbers.';
  fGroupNames[4].Name := 'Comments';
  fGroupNames[4].Desc := 'All types of commentaries //..., {...}, (*...*).';
  fGroupNames[5].Name := 'Symbols';
  fGroupNames[5].Desc := 'All special symbols.';
  fGroupNames[6].Name := 'Assembler';
  fGroupNames[6].Desc := 'Blocks of code with inline assembler.';
  fGroupNames[7].Name := 'Directives';
  fGroupNames[7].Desc := 'Compiler directives and comditional defines.';
end;

{ TOYDelphi5TemplatedParser }
constructor TOYDelphi5TemplatedParser.Create(const Url : String; const IConsumer : IOYLinguineConsumer);
begin
  inherited Create(Url);
  fIConsumer := IConsumer;

  fPreproc := TOYDelphi5Preproc.Create(Url);
  fParsed := False;
end;                                                                              

destructor TOYDelphi5TemplatedParser.Destroy;
begin
  fPreproc.Free;
  inherited Destroy;
end;

function TOYDelphi5TemplatedParser.yylex : TOYParserToken;

   procedure _PushToken(Token : TOYParserToken);
   begin
     if fNeedOutput then begin
       FIConsumer.BeginElementWithAttrs(
         cTokenElementName,
         cElementIdAttr, IntToStr(Token.Frame.Id),
         cElementLineNumAttr, IntToStr(Token.Frame.leftLineIndex)
       );
       try
         FIConsumer.AddValue(textByToken(Token));
       finally
         FIConsumer.EndElement(cTokenElementName);
       end;
     end;
   end;

   procedure _PushPendingJunk(var S :String);
   begin
     if fNeedOutput then begin
       if (S <> '') then begin
         FIConsumer.BeginElementWithAttr(cTokenElementName, cElementIdAttr, IntToStr(tkJunk));
         try
           FIConsumer.AddValue(S);
           S := '';
         finally
           FIConsumer.EndElement(cTokenElementName);
         end;
       end;
     end;
   end;               

   function _ShouldSkipToken(Token : TOYParserToken; var S : String) : Boolean;
   begin
     Result := True;
     case Token.Frame.Id of
       tkCompiler, tkAnsiComment, tkBorComment, tkSlashesComment : begin
         _PushPendingJunk(S);
         _PushToken(Token);
       end;
       tkUnknown : begin
         if Assigned(OnHint) then begin
           if Length(textByToken(Token)) = 1 then begin
             OnHint(urlByToken(Token), 'Unknown token (ASCII code ' + IntToStr(Ord(textByToken(Token)[1])) + ', replaced with "?").', lineByToken(Token), TextByToken(Token), Token.Frame.leftLineIndex, Token.Frame.leftColIndex);
           end else begin
             OnHint(urlByToken(Token), 'Unknown token (text "' + textByToken(Token) + '", replaced with "?").', lineByToken(Token), TextByToken(Token), Token.Frame.leftLineIndex, Token.Frame.leftColIndex);
           end;
         end;
         if OYAllCharsInRange(textByToken(Token), cValidXmlValueChars) then begin
           S := S + textByToken(Token);
         end else begin
           if fNeedOutput then begin
             FIConsumer.BeginElementWithAttr(cTokenElementName, cElementIdAttr, IntToStr(Token.Frame.Id));
             try
               FIConsumer.AddValue('?');
             finally
               FIConsumer.EndElement(cTokenElementName);
             end;
           end;                       
         end;             
       end;
       tkCRLF, tkSpace, tkIgnore: begin
         S := S + textByToken(Token);
       end;
       else Result := False;
     end;
   end;

var
  S : String;
begin
  S := '';
  repeat
    Result := fPreproc.Lex;
  until (Result.Lexer = nil) or not _ShouldSkipToken(Result, S);

  _PushPendingJunk(S);
  if Result.Lexer <> nil then begin
    Result.Index := fTokenIndex;
    inc(fTokenIndex);
  end;
end;

function TOYDelphi5TemplatedParser.yyparse : Boolean;
begin
  fTokenIndex := 0;
  fNeedOutput := False;

  fPreproc.OnWarning := OnWarning;
  fPreproc.OnHint := OnHint;
  fPreproc.OnCanContinue := OnCanContinue;

  Result := inherited yyparse;
  fParsed := Result;
end;

procedure TOYDelphi5TemplatedParser.yyproduce;
begin
  if not fParsed then begin
    raise Exception.Create('Unable to produce output(parsing incomplete).');
  end;
  fTokenIndex := 0;
  fNeedOutput := True;

  TransformFrames;
  SortFrames;

  fPreproc.FirstToken;
  WriteToConsumer(FIConsumer);
end;

function TOYDelphi5TemplatedParser.currentToken : POYParserToken;
begin
  fPreproc.PrevToken;
  Result := fPreproc.CurrentToken;
  fPreproc.NextToken;
end;

function TOYDelphi5TemplatedParser.lineByToken(Token : TOYParserToken) : String;
begin
  Result := fPreproc.lineByToken(Token);
end;

function TOYDelphi5TemplatedParser.textByToken(Token : TOYParserToken) : String;
begin
  Result := fPreproc.textByToken(Token);
end;

function TOYDelphi5TemplatedParser.urlByToken(Token : TOYParserToken) : String;
begin
  Result := fPreproc.urlByToken(Token);
end;

end.
