{$I oycl.inc}  

{
  Turbo Pascal Lex and Yacc Version 4.1
  Compiler generator toolset for Turbo/Borland/Free Pascal and Delphi

  Copyright (c) 1990-2000 by Albert Graef
                        <ag@muwiinfa.geschichte.uni-mainz.de>
  Copyright (c) 1996 by Berend de Boer
                        <berend@pobox.com>
  Copyright (c) 1998 by Michael Van Canneyt
                        <Michael.VanCanneyt@fys.kuleuven.ac.be>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

}

unit LexLib;

interface

uses { vcl } Classes, SysUtils,
     { oy } OYParserToken;

const
  max_chars   = 64 * 1024;
  max_text    = 64 * 1024;
  max_matches =  1 * 1024;
  max_rules   =       256;

type
  TLexer = class(TObject)
  protected
    yystate    : Integer; (* current state of lexical analyzer *)
    yyactchar  : Char;    (* current character *)
    yylastchar : Char;    (* last matched character (#0 if none) *)
    yyrule     : Integer; (* matched rule *)
    yyrejected : Boolean; (* current match rejected? *)
    yydone     : Boolean; (* yylex return value set? *)
    yyretval   : Integer; (* yylex return value *)
    bufptr     : Integer; (* position in current line buffer *)
    buf        : array [1..max_chars] of Char;  (* current line buffer *)
    yystext            : String;
    yysstate, yylstate : Integer;
    yymatches          : Integer;
    yystack            : array [1..max_matches] of Integer;
    yypos              : array [1..max_rules] of Integer;
    yysleng            : Byte;
    yyline             : String;         (* current input line *)
    yylineno, yycolno  : Integer;        (* current input position *)
    yytext             : String;         (* matched text (should be considered r/o) *)
    fUrl               : String;         (* url of file being lexed *)
    fInput             : TStrings;       (* string list that hold the input file *)
    fFrame             : TOYLexerFrame;  (* frame for token returned by last call to lex *)
    fOnWarning : TOYParseEvent;
    fOnHint : TOYParseEvent;
    fOnCanContinue : TOYParserCanContinue;

  protected
    function get_char : Char;
    procedure unget_char ( c : Char );
    procedure yyshiftFrame;
    procedure yystart ( state : Integer );
    procedure yymore;
    procedure yyless ( n : Integer );
    procedure yyreject;
    procedure yyreturn ( n : Integer );
    procedure yyreturnc ( c : Char );
    function yywrap : Boolean;
    procedure yynew;
    procedure yyscan;
    procedure yymark ( n : Integer );
    procedure yymatch ( n : Integer );
    function yyfind ( var n : Integer ) : Boolean;
    function yydefault : Boolean;
    procedure yyclear;

    procedure yywarning ( msg : String );
    procedure yyhint ( msg : String );
    function yycancontinue : Boolean;

  public
    constructor Create(const url : String);
    destructor Destroy; override;

  public
    function lex : Integer; virtual; abstract;
    function frame : TOYLexerFrame;
    function url : String;
    function lineByFrame(Frame : TOYLexerFrame) : String;
    function textByFrame(Frame : TOYLexerFrame) : String;

  public
    property OnWarning : TOYParseEvent read fOnWarning write fOnWarning;
    property OnHint : TOYParseEvent read fOnHint write fOnHint;
    property OnCanContinue : TOYParserCanContinue read fOnCanContinue write fOnCanContinue;

  end;

implementation

const nl = #10;  (* newline character *)

constructor TLexer.Create(const url : String);
begin
  inherited Create;

  fUrl := url;

  fInput := TStringList.Create;
  fInput.LoadFromFile(URL);

  yyclear;
end;(*create*)

destructor TLexer.Destroy;
begin
  fInput.Free;
  inherited Destroy;
end;(*destroy*)

procedure TLexer.yywarning ( msg : String );
begin
  if Assigned(fOnWarning) then begin
    fOnWarning(fUrl, msg, '', '', -1, -1);
  end;
end;

procedure TLexer.yyhint ( msg : String );
begin
  if Assigned(fOnHint) then begin
    fOnHint(fUrl, msg, '', '', -1, -1);
  end;
end;

function TLexer.yycancontinue : Boolean;
begin
  Result := True;
  if Assigned(fOnCanContinue) then begin
    Result := fOnCanContinue;
  end;
end;

function TLexer.get_char : Char;
var
  i : Integer;
begin
  if (bufptr=0) and (yylineno < fInput.Count) then
    begin
      yyline := fInput[yylineno];
      inc(yylineno);
      if ((yylineno mod 100) = 0) and (not yycancontinue) then begin
        raise Exception.Create('Processing aborted.');
      end;
      fFrame.rightLineIndex := yylineno;
      yycolno := 1;
      fFrame.rightColIndex := yycolno;
      buf[1] := nl;
      for i := 1 to length(yyline) do
        buf[i+1] := yyline[length(yyline)-i+1];
      inc(bufptr, length(yyline)+1);
    end;
  if bufptr>0 then
    begin
      get_char := buf[bufptr];
      dec(bufptr);
      inc(yycolno);
      fFrame.rightColIndex := yycolno;
    end
  else
    get_char := #0;
end(*get_char*);

procedure TLexer.unget_char ( c : Char );
begin
  if bufptr=max_chars then begin
    raise Exception.Create('Input buffer overflow.');
  end;
  inc(bufptr);
  dec(yycolno);
  fFrame.rightColIndex := yycolno;
  buf[bufptr] := c;
end(*unget_char*);

procedure TLexer.yymore;
begin
  yystext := yytext;
end(*yymore*);

procedure TLexer.yyless ( n : Integer );
var i : Integer;
begin
  for i := Length(yytext) downto n+1 do
    unget_char(yytext[i]);
  SetLength(yytext, n);
end(*yyless*);

procedure TLexer.yyreject;
var i : Integer;
begin
  yyrejected := true;
  for i := Length(yytext)+1 to yysleng do
    yytext := yytext+get_char;
  dec(yymatches);
end(*reject*);

procedure TLexer.yyreturn ( n : Integer );
begin
  yyretval := n;
  yydone := true;
end(*return*);

procedure TLexer.yyreturnc ( c : Char );
begin
  yyretval := ord(c);
  yydone := true;
end(*returnc*);

procedure TLexer.yystart ( state : Integer );
begin
  yysstate := state;
end(*start*);                        

function TLexer.yywrap : Boolean;
begin
  yywrap := true;
end(*yywrap*);

procedure TLexer.yynew;
begin
  if yylastchar<>#0 then
    if yylastchar=nl then
      yylstate := 1
    else
      yylstate := 0;
  yystate := yysstate+yylstate;
  yytext  := yystext;
  yystext := '';
  yymatches := 0;
  yydone := false;
end(*yynew*);

procedure TLexer.yyscan;
begin
  if Length(yytext)>max_text then begin
    raise Exception.Create('Line is too long.');
  end;
  yyactchar := get_char;
  yytext := yytext + yyactchar;
end(*yyscan*);

procedure TLexer.yymark ( n : Integer );
begin
  if n>max_rules then begin
    raise Exception.Create('Too many rules.');
  end;
  yypos[n] := Length(yytext);
end(*yymark*);

procedure TLexer.yymatch ( n : Integer );
begin
  inc(yymatches);
  if yymatches>max_matches then begin
    raise Exception.Create('Match stack overflow.');
  end;
  yystack[yymatches] := n;
end(*yymatch*);

function TLexer.yyfind ( var n : Integer ) : Boolean;
begin
  yyrejected := false;
  while (yymatches>0) and (yypos[yystack[yymatches]]=0) do
    dec(yymatches);
  if yymatches>0 then
    begin
      yysleng := Length(yytext);
      n       := yystack[yymatches];
      yyless(yypos[n]);
      yypos[n] := 0;
      if Length(yytext)>0 then
        yylastchar := yytext[Length(yytext)]
      else
        yylastchar := #0;
      yyfind := true;
    end
  else
    begin
      yyless(0);
      yylastchar := #0;
      yyfind := false;
    end
end(*yyfind*);

function TLexer.yydefault : Boolean;
begin
  yyrejected := false;
  yyactchar := get_char;
  if yyactchar<>#0 then
    begin
      yydefault := true;
    end
  else
    begin
      yylstate := 1;
      yydefault := false;
    end;
  yylastchar := yyactchar;
end(*yydefault*);

procedure TLexer.yyclear;
begin
  bufptr := 0;
  yysstate := 0;
  yylstate := 1;
  yylastchar := #0;
  yytext := '';
  yystext := '';
  
  fFrame.leftLineIndex := -1;
  fFrame.leftColIndex := 0;
  fFrame.rightLineIndex := -1;
  fFrame.rightColIndex := 0;

end(*yyclear*);

procedure TLexer.yyshiftFrame;
begin
  { shift frame left side indexes }
  if (fFrame.rightColIndex = -1) or (fFrame.rightLineIndex = -1) then begin
    { init frame the first time }
    fFrame.leftLineIndex := 1;
    fFrame.leftColIndex := 1;
    fFrame.rightLineIndex := 1;
    fFrame.rightColIndex := 1;
  end else begin
    fFrame.leftLineIndex := fFrame.rightLineIndex;
    fFrame.leftColIndex := fFrame.rightColIndex;
    if bufptr = 0 then begin
      fFrame.leftLineIndex := fFrame.leftLineIndex + 1;
      fFrame.leftColIndex := 1;
    end;
  end;
end;

function TLexer.url : String;
begin
  Result := fUrl;
end;(*url*)

function TLexer.frame : TOYLexerFrame;
begin
  Result := fFrame;
end;

function TLexer.lineByFrame(Frame : TOYLexerFrame) : String;
var
  ll : Integer;
begin
  ll := Frame.leftLineIndex - 1;
  if (ll >= fInput.Count) then begin
    Result := '';
  end else begin
    Result := fInput[ll];
  end;
end;(*lineByIndex*)

function TLexer.textByFrame(Frame : TOYLexerFrame) : String;
var
  i : Integer;
  ll, rl, lc, rc : Integer;
begin
  ll := Frame.leftLineIndex - 1;
  lc := Frame.leftColIndex;
  rl := Frame.rightLineIndex - 1;                  
  rc := Frame.rightColIndex;
  if (ll >= fInput.Count) or (rl >= fInput.Count) then begin
    Result:= '';
  end else begin
    if (ll = rl) then begin
      Result := Copy(fInput[ll], lc, rc - lc);
    end else begin
      if (ll < rl) then begin
        Result := Copy(fInput[ll], lc, Length(fInput[ll]));
        for i := ll + 1 to rl - 1 do begin
          if i = ll + 1 then begin
            Result := Result + #$D#$A;
          end;                                          
          Result := Result + fInput[i] + #$D#$A;
        end;
        Result := Result + Copy(fInput[rl], 1, rc - 1);
      end else begin                                                         
        raise Exception.Create('Invalid left/right index in token frame.');
      end;
    end;
  end;
end;(*textByFrame*)

end(*LexLib*).
